/*
 * Copyright (C) 2017 Open Source Robotics Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
*/

//                ****** Do not modify this file. ******
// This file is automatically generated by CMake. Changes should instead be
// made to cmake/ign_auto_headers.hh.in in ignition-cmake

#include <ignition/transport/config.hh>
#include <ignition/transport/AdvertiseOptions.hh>
#include <ignition/transport/CIface.h>
#include <ignition/transport/Clock.hh>
#include <ignition/transport/Discovery.hh>
#include <ignition/transport/HandlerStorage.hh>
#include <ignition/transport/Helpers.hh>
#include <ignition/transport/MessageInfo.hh>
#include <ignition/transport/NetUtils.hh>
#include <ignition/transport/Node.hh>
#include <ignition/transport/NodeOptions.hh>
#include <ignition/transport/NodeShared.hh>
#include <ignition/transport/Packet.hh>
#include <ignition/transport/Publisher.hh>
#include <ignition/transport/RepHandler.hh>
#include <ignition/transport/ReqHandler.hh>
#include <ignition/transport/SubscribeOptions.hh>
#include <ignition/transport/SubscriptionHandler.hh>
#include <ignition/transport/TopicStorage.hh>
#include <ignition/transport/TopicUtils.hh>
#include <ignition/transport/TransportTypes.hh>
#include <ignition/transport/Uuid.hh>

